"""base resource and resource root factory"""

import abc

from pyramid.authorization import DENY_ALL, Allow, Authenticated, Everyone


class BaseResource(abc.ABC):
    """Base Resource for all other resources"""

    __parent__ = None  # required by pyramid for location aware resources
    __name__ = None  # required by pyramid for location aware resources

    request = None  # current request object
    model = None  # a related sqlalchemy model
    nodes = {}  # static child nodes

    def __init__(self, name, parent, sql_model_instance=None):
        self.__name__ = name
        self.__parent__ = parent
        self.request = parent.request
        self.model = sql_model_instance

    @abc.abstractmethod
    def __acl__(self):
        """Access controll list"""

    def __getitem__(self, key):
        """returns child resources"""
        child_node_class = self.nodes[key]
        return child_node_class(key, self)


class User(BaseResource):
    def __acl__(self):
        """access controll list"""
        acl = [(Allow, "role:admin", "view"), (Allow, "role:admin", "edit")]
        if not self.model.is_active:
            acl.append((Allow, "role:admin", "delete"))
        acl.append(DENY_ALL)
        return acl

    @classmethod
    def from_model(cls, model, parent):
        """initializes a resource from an SQLalchemy object"""
        return cls(model.username, parent, model)


class UserList(BaseResource):
    def __acl__(self):
        """access controll list"""
        return [(Allow, "role:admin", "view"), DENY_ALL]

    def __getitem__(self, key):
        """returns child resources"""
        try:
            user = self.request.repo.get_user_by_username(key)
            return User.from_model(user, self)
        except StopIteration as e:
            raise KeyError from e


class Order(BaseResource):
    def __acl__(self):
        """access controll list"""
        acl = [
            (Allow, "role:user", "view"),
            (Allow, "role:purchaser", "edit"),
            (Allow, "role:purchaser", "delete"),
        ]
        if self.model.in_process:
            (acl.append((Allow, "role:purchaser", "reorder")),)
            (acl.append((Allow, "role:user", "reorder")),)
        else:
            acl.append((Allow, f"user:{self.model.created_by}", "edit"))
            acl.append((Allow, f"user:{self.model.created_by}", "delete"))
        acl.append(DENY_ALL)
        return acl

    @classmethod
    def from_model(cls, model, parent):
        return cls(model.id, parent, model)


class OrderList(BaseResource):
    def __acl__(self):
        """access controll list"""
        return [
            (Allow, "role:user", "view"),
            (Allow, "role:user", "add"),
            (Allow, "role:purchaser", "batch-edit"),
            (Allow, "role:purchaser", "batch-delete"),
            DENY_ALL,
        ]

    def __getitem__(self, key):
        """returns child resources"""
        try:
            order = self.request.repo.get_order(key)
            return Order.from_model(order, self)
        except StopIteration as e:
            raise KeyError from e


class Vendor(BaseResource):
    def __acl__(self):
        """access controll list"""
        acl = [
            (Allow, "role:admin", "view"),
            (Allow, "role:admin", "edit"),
            DENY_ALL,
        ]
        return acl

    @classmethod
    def from_model(cls, model, parent):
        """initializes a resource from an model object"""
        return cls(model.name, parent, model)


class VendorList(BaseResource):
    def __acl__(self):
        """access controll list"""
        return [(Allow, "role:admin", "view"), DENY_ALL]

    def __getitem__(self, key):
        """returns child resources"""
        try:
            aggregagtes = self.request.repo.get_vendor_aggregates(key)
            return Vendor.from_model(aggregagtes, self)
        except StopIteration as e:
            raise KeyError from e


class Root(BaseResource):
    """Root resource"""

    __name__ = None
    __parent__ = None

    nodes = {"users": UserList, "orders": OrderList, "vendors": VendorList}

    def __init__(self, request):
        self.request = request

    def __acl__(self):
        """access controll list"""
        return [
            (Allow, Everyone, "login"),
            (Allow, Everyone, "logout"),
            (Allow, Everyone, "registration"),
            (Allow, Everyone, "view"),
            (Allow, Authenticated, "account"),
        ]


def includeme(config):
    """
    Initialize the resources for traversal in a Pyramid app.

    Activate this setup using ``config.include('ordr3.resources')``.

    """
    config.set_root_factory(Root)
