import deform
from pyramid.csrf import get_csrf_token
from pyramid.httpexceptions import HTTPFound
from pyramid.view import view_config
from sqlalchemy import or_  # , func

from .. import events, models, resources, services
from ..schemas import orders
from . import DefaultQueryParams, get_offset

QUERY_LIMIT = 25


def get_status(request):
    status_param = request.GET.get("status", "")
    try:
        return models.OrderStatus[status_param.upper()]
    except KeyError:
        return None


def get_category(request):
    category_param = request.GET.get("category", "")
    try:
        return models.OrderCategory[category_param.upper()]
    except KeyError:
        return None


def get_multiple_orders(context, request):  # noqa: ARG001
    order_ids = {v for k, v in request.POST.items() if k == "selection"}
    return (
        request.repo.session.query(models.OrderItem)
        .filter(models.OrderItem.id.in_(order_ids))
        .order_by(models.OrderItem.created_on.desc())
        .all()
    )


def get_form_data_from_order(order):
    status = {"status": order.status.name}
    item = {
        "cas_description": order.cas_description,
        "category": order.category.name,
        "vendor": order.vendor,
        "catalog_nr": order.catalog_nr,
        "package_size": order.package_size,
    }
    pricing = {
        "unit_price": {
            "amount": "%.2f" % order.unit_price,
            "currency": order.currency,
        },
        "quantity": order.amount,
        "total_price": {
            "amount": "%.2f" % order.total_price,
            "currency": order.currency,
        },
    }

    funding = {"funds": order.account}

    optional = {"comment": order.comment}

    form_data = {
        "status": status,
        "item": item,
        "pricing": pricing,
        "funding": funding,
        "optional": optional,
    }
    return form_data


def update_order_with_form_data(order, form_data):
    form_item = form_data["item"]
    form_pricing = form_data["pricing"]
    form_funding = form_data["funding"]
    form_optional = form_data["optional"]

    order.cas_description = form_item["cas_description"]
    order.category = models.OrderCategory[form_item["category"]]
    order.vendor = form_item["vendor"]
    order.catalog_nr = form_item["catalog_nr"]
    order.package_size = form_item["package_size"]

    order.unit_price = form_pricing["unit_price"]["amount"]
    order.currency = form_pricing["unit_price"]["currency"]
    order.amount = form_pricing["quantity"]

    order.account = form_funding["funds"]

    order.comment = form_optional["comment"]


@view_config(
    context="ordr3:resources.OrderList",
    permission="view",
    request_method="GET",
    renderer="ordr3:templates/orders/list.jinja2",
)
@view_config(
    context="ordr3:resources.OrderList",
    permission="view",
    request_method="GET",
    xhr=True,
    renderer="ordr3:templates/orders/list_content.jinja2",
)
def order_list(context, request):
    limit = QUERY_LIMIT
    offset = get_offset(request)
    status = get_status(request)
    category = get_category(request)
    username = request.GET.get("user", None)
    search = request.GET.get("search", None)

    query = request.repo.session.query(models.OrderItem).order_by(
        models.OrderItem.created_on.desc()
    )
    if status:
        query = query.filter(models.OrderItem.status == status)
    if category:
        query = query.filter(models.OrderItem.category == category)
    if username == "-purchaser-":
        query = (
            query.distinct()
            .join(models.OrderItem.log)
            .filter(models.LogEntry.by == request.user.username)
        )
    elif username:
        query = query.filter(models.OrderItem.created_by == username)
    if search:
        term = f"%{search}%"
        query = query.filter(
            or_(
                models.OrderItem.cas_description.ilike(term),
                models.OrderItem.vendor.ilike(term),
                models.OrderItem.catalog_nr.ilike(term),
                models.OrderItem.account.ilike(term),
                models.OrderItem.created_by.ilike(term),
            )
        )
    orders = query[offset : offset + limit]

    next_offset = None if limit != len(orders) else (offset + limit)
    order_resources = [resources.Order.from_model(u, context) for u in orders]

    filter_status = "all" if status is None else status.name.lower()
    filter_category = None if category is None else category.name.lower()
    query_defaults = DefaultQueryParams(
        status=filter_status,
        user=username,
        search=search,
        category=filter_category,
    )

    return {
        "is_order_list": True,
        "next_offset": next_offset,
        "orders": order_resources,
        "stati": models.OrderStatus,
        "query_defaults": query_defaults,
    }


@view_config(
    context="ordr3:resources.OrderList",
    name="batch-delete",
    permission="batch-delete",
    request_method="POST",
    require_csrf=False,
    renderer="ordr3:templates/orders/batch_delete.jinja2",
)
def batch_delete(context, request):
    submitted_orders = get_multiple_orders(context, request)
    orders = [o for o in submitted_orders if not o.in_process]

    if not orders:
        return HTTPFound(request.resource_url(context))

    if len(orders) > QUERY_LIMIT:
        request.emit(
            events.FlashMessage.warning(
                f"You cannot delete more than "
                f"{QUERY_LIMIT} orders in one sweep."
            )
        )
        return HTTPFound(request.resource_url(context))

    return {"orders": orders, "csrf_token": get_csrf_token(request)}


@view_config(
    context="ordr3:resources.OrderList",
    name="batch-delete-confirm",
    permission="batch-delete",
    request_method="POST",
)
def batch_delete_confirmed(context, request):
    if "delete" not in request.POST:
        return HTTPFound(request.resource_url(context.__parent__))
    if request.POST.get("confirmation", "") != "confirmed":
        return HTTPFound(request.resource_url(context.__parent__))

    submitted_orders = get_multiple_orders(context, request)
    orders = [o for o in submitted_orders if not o.in_process]

    if not orders:
        return HTTPFound(request.resource_url(context))

    if len(orders) > QUERY_LIMIT:
        request.emit(
            events.FlashMessage.warning(
                f"You cannot delete more than "
                f"{QUERY_LIMIT} orders in one sweep."
            )
        )
    else:
        for order in orders:
            request.repo.delete_order(order)
        request.emit(
            events.FlashMessage.info(f"{len(orders)} orders have been deleted.")
        )
    return HTTPFound(request.resource_url(context))


@view_config(
    context="ordr3:resources.OrderList",
    name="batch-edit",
    permission="batch-edit",
    request_method="POST",
    require_csrf=False,
    renderer="ordr3:templates/orders/batch_edit.jinja2",
)
def batch_edit(context, request):
    orders = get_multiple_orders(context, request)

    if not orders:
        return HTTPFound(request.resource_url(context))

    if len(orders) > QUERY_LIMIT:
        request.emit(
            events.FlashMessage.warning(
                f"You cannot edit more than {QUERY_LIMIT} orders in one sweep."
            )
        )
        return HTTPFound(request.resource_url(context))

    return {
        "orders": orders,
        "stati": models.OrderStatus,
        "csrf_token": get_csrf_token(request),
    }


@view_config(
    context="ordr3:resources.OrderList",
    name="batch-edit-confirm",
    permission="batch-edit",
    request_method="POST",
    require_csrf=False,
)
def batch_edit_confirm(context, request):
    if "change" not in request.POST:
        return HTTPFound(request.resource_url(context.__parent__))

    orders = get_multiple_orders(context, request)

    if not orders:
        return HTTPFound(request.resource_url(context))

    if len(orders) > QUERY_LIMIT:
        request.emit(
            events.FlashMessage.warning(
                f"You cannot edit more than {QUERY_LIMIT} orders in one sweep."
            )
        )
        return HTTPFound(request.resource_url(context))

    changes = {}

    for order in orders:
        status_name = request.POST.get(f"status-{order.id}")
        try:
            status = models.OrderStatus[status_name.upper()]
            is_noteworthy = services.create_log_entry(
                order, status, request.user
            )
            if is_noteworthy:
                if order.created_by not in changes:
                    changes[order.created_by] = []
                changes[order.created_by].append(order)
        except KeyError:
            pass

    for username, user_orders in changes.items():
        user = request.repo.get_user_by_username(username)
        request.emit(events.OrderStatusChangeEmail(user, user_orders))

    if len(orders):
        request.emit(
            events.FlashMessage.info(f"{len(orders)} orders have been updated.")
        )

    return HTTPFound(request.resource_url(context))


@view_config(
    context="ordr3:resources.OrderList",
    name="vendor",
    permission="add",
    request_method="POST",
    xhr=True,
    require_csrf=False,
    renderer="json",
)
def check_vendor_name(context, request):  # noqa: ARG001
    vendor = request.POST.get("vendor", "")
    result = services.check_vendor_name(request.repo, vendor)
    return result._asdict()


@view_config(
    context="ordr3:resources.OrderList",
    name="add",
    permission="add",
    request_method="GET",
    renderer="ordr3:templates/orders/add.jinja2",
)
def new_order(context, request):
    autocorrect_url = request.resource_url(context, "vendor")
    form = orders.AddOrderSchema.as_form(
        request, autocorrect_url=autocorrect_url
    )
    consumables = services.find_consumables(request.repo)
    return {
        "form": form,
        "categories": models.OrderCategory,
        "consumables": consumables,
    }


@view_config(
    context="ordr3:resources.OrderList",
    name="add",
    permission="add",
    request_method="POST",
    renderer="ordr3:templates/orders/add.jinja2",
)
def place_new_order(context, request):
    """process the reorder form"""
    autocorrect_url = request.resource_url(context, "vendor")
    form = orders.AddOrderSchema.as_form(
        request, autocorrect_url=autocorrect_url
    )

    if "Place_Order" not in request.POST:
        return HTTPFound(request.resource_url(context.__parent__))

    data = request.POST.items()

    try:
        appstruct = form.validate(data)
    except deform.ValidationFailure:
        consumables = services.find_consumables(request.repo)
        return {
            "form": form,
            "categories": models.OrderCategory,
            "consumables": consumables,
        }

    # form validation sucessful, change order
    order = models.OrderItem(None, None, None, None, None, None, None, None)
    update_order_with_form_data(order, appstruct)

    services.create_log_entry(order, models.OrderStatus.OPEN, request.user)

    request.repo.add_order(order)

    request.emit(
        events.FlashMessage.info(
            f"The order {order.cas_description} has been placed."
        )
    )

    return HTTPFound(request.resource_url(context.__parent__))


@view_config(
    context="ordr3:resources.Order",
    permission="view",
    name="view",
    request_method="GET",
    renderer="ordr3:templates/orders/view.jinja2",
)
def view_order(context, request):  # noqa: ARG001
    return {"csrf_token": get_csrf_token(request)}


@view_config(
    context="ordr3:resources.Order",
    name="edit",
    permission="edit",
    request_method="GET",
    renderer="ordr3:templates/orders/edit.jinja2",
)
def edit_order(context, request):
    autocorrect_url = request.resource_url(context.__parent__, "vendor")
    form = orders.EditOrderSchema.as_form(
        request, autocorrect_url=autocorrect_url
    )

    form_data = get_form_data_from_order(context.model)
    form.set_appstruct(form_data)

    return {"form": form}


@view_config(
    context="ordr3:resources.Order",
    name="edit",
    permission="edit",
    request_method="POST",
    renderer="ordr3:templates/orders/edit.jinja2",
)
def do_edit_order(context, request):
    """process the edit order form"""
    autocorrect_url = request.resource_url(context.__parent__, "vendor")
    form = orders.EditOrderSchema.as_form(
        request, autocorrect_url=autocorrect_url
    )

    if "Save_Changes" not in request.POST:
        return HTTPFound(request.resource_url(context.__parent__))

    data = request.POST.items()

    try:
        appstruct = form.validate(data)
    except deform.ValidationFailure:
        return {"form": form}

    # form validation sucessful, change order
    order = context.model
    update_order_with_form_data(order, appstruct)
    form_status_section = appstruct["status"]

    status = models.OrderStatus[form_status_section["status"]]
    is_noteworthy = services.create_log_entry(order, status, request.user)

    if is_noteworthy:
        user = request.repo.get_user_by_username(order.created_by)
        request.emit(events.OrderStatusChangeEmail(user, [order]))

    request.emit(
        events.FlashMessage.info(
            f"The order {context.model.cas_description} has been changed."
        )
    )

    return HTTPFound(request.resource_url(context.__parent__))


@view_config(
    context="ordr3:resources.Order",
    name="reorder",
    permission="reorder",
    request_method="GET",
    renderer="ordr3:templates/orders/reorder.jinja2",
)
def reorder(context, request):
    autocorrect_url = request.resource_url(context.__parent__, "vendor")
    form = orders.AddOrderSchema.as_form(
        request, autocorrect_url=autocorrect_url
    )

    form_data = get_form_data_from_order(context.model)
    form.set_appstruct(form_data)

    return {"form": form}


@view_config(
    context="ordr3:resources.Order",
    name="reorder",
    permission="reorder",
    request_method="POST",
    renderer="ordr3:templates/orders/reorder.jinja2",
)
def place_reorder(context, request):
    """process the reorder form"""
    autocorrect_url = request.resource_url(context.__parent__, "vendor")
    form = orders.AddOrderSchema.as_form(
        request, autocorrect_url=autocorrect_url
    )

    if "Place_Order" not in request.POST:
        return HTTPFound(request.resource_url(context.__parent__))

    data = request.POST.items()

    try:
        appstruct = form.validate(data)
    except deform.ValidationFailure:
        return {"form": form}

    # form validation sucessful, change order
    order = models.OrderItem(None, None, None, None, None, None, None, None)
    update_order_with_form_data(order, appstruct)

    services.create_log_entry(order, models.OrderStatus.OPEN, request.user)

    request.repo.add_order(order)

    request.emit(
        events.FlashMessage.info(
            f"The order {order.cas_description} has been placed."
        )
    )

    return HTTPFound(request.resource_url(context.__parent__))


@view_config(
    context="ordr3:resources.Order",
    name="delete",
    permission="delete",
    request_method="GET",
    renderer="ordr3:templates/orders/delete.jinja2",
)
def delete_order(context, request):  # noqa: ARG001
    return {"csrf_token": get_csrf_token(request)}


@view_config(
    context="ordr3:resources.Order",
    name="delete",
    permission="delete",
    request_method="POST",
)
def delete_confirmed(context, request):
    if "delete" not in request.POST:
        return HTTPFound(request.resource_url(context.__parent__))
    if request.POST.get("confirmation", "") != "confirmed":
        return HTTPFound(request.resource_url(context.__parent__))

    request.emit(
        events.FlashMessage.info(
            f"The order {context.model.cas_description} has been deleted."
        )
    )
    request.repo.delete_order(context.model)

    return HTTPFound(request.resource_url(context.__parent__))
